import os
import sys
import subprocess

# returns in_path,gpg_path
def find_gpg():
	if sys.platform == 'win32':
		in_path,gpgpath = find_gpg_windows()
		# Removed due to gnupg.py v0.37
		#if gpgpath != None and gpgpath.find(' ') >= 0:
		#	gpgpath = '"' + gpgpath + '"'
	elif sys.platform == 'darwin':
		in_path,gpgpath = find_gpg_macos()
	else:
		in_path,gpgpath = find_gpg_unix()
	return in_path,gpgpath

def find_gpg_windows():
	try:
		for line in os.environ['PATH'].split(os.pathsep):
			gpgpath = line + os.sep + 'gpg.exe'
			if os.path.isfile(gpgpath):
				return True,gpgpath
	except Exception:
		pass

	try:
		# Look for local copy in the install directory first
		program_path = os.path.dirname(os.path.realpath(sys.argv[0]))
		program_path_bin = program_path + os.sep + 'bin'

		#for line in [ "C:\\Program Files (x86)\\GNU\\GnuPG.v2\\bin",program_path,'C:\\Program Files (x86)\\GNU\\GnuPG','C:\\Program Files\\GNU\\GnuPG',program_path_bin,'C:\\Program Files (x86)\\GNU\\GnuPG\\bin','C:\\Program Files\\GNU\\GnuPG\\bin' ]:
		for line in [ program_path,'C:\\Program Files (x86)\\GNU\\GnuPG','C:\\Program Files\\GNU\\GnuPG',program_path_bin,'C:\\Program Files (x86)\\GNU\\GnuPG\\bin','C:\\Program Files\\GNU\\GnuPG\\bin' ]:
			gpgpath = line + os.sep + 'gpg.exe'
			if os.path.isfile(gpgpath):
				return False,gpgpath
	except Exception:
		pass

	return False,None

def find_gpg_macos():
	try:
		program_path = os.path.dirname(os.path.realpath(sys.argv[0]))
		bin_path = program_path + os.sep + 'bin'
		search_paths = [ program_path,bin_path,'_NOT_IN_PATH_' ]
		search_paths.append(os.environ['PATH'].split(os.pathsep))
		inpath = True
		for line in search_paths:
			if line == '_NOT_IN_PATH_':
				inpath = False
			gpgpath = line + os.sep + 'gpg'
			if os.path.isfile(gpgpath):
				return inpath,gpgpath
	except Exception:
		pass
	return False,None

def find_gpg_unix():
	try:
		for line in os.environ['PATH'].split(os.pathsep):
			gpgpath = line + os.sep + 'gpg'
			if os.path.isfile(gpgpath):
				return True,gpgpath
	except Exception:
		pass
	return False,None

# This is the homedir for Confidant Mail files. It is unrelated to GPG.
def find_default_homedir():
	appdata = None
	if sys.platform == 'win32':
		varname = 'APPDATA'
	else:
		varname = 'HOME'
	if varname in os.environ:
		basedir = os.environ[varname]
	if basedir != None and os.path.exists(basedir):
		homedir = basedir + os.sep + 'confidantmail'
		if os.path.exists(homedir):
			return homedir
		else:
			try:
				os.mkdir(homedir)
			except Exception:
				pass
			if os.path.exists(homedir):
				return homedir
			else:
				return None
	else:
		return None
			
def is_gpgv2(gpgpath):
	try:
		cmdline = [ gpgpath,"--version" ]
		if sys.platform == 'win32':
			# http://stackoverflow.com/questions/7006238/how-do-i-hide-the-console-when-i-use-os-system-or-subprocess-call/7006424#7006424
			CREATE_NO_WINDOW = 0x08000000
			cmdh = subprocess.Popen(cmdline,bufsize=16384,stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,shell = False,creationflags = CREATE_NO_WINDOW)
		else:
			cmdh = subprocess.Popen(cmdline,bufsize=16384,stdin=subprocess.PIPE,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,shell = False)
		cmdh.stdin.close() # undefined stdin -> Windows fails
		gpgver = cmdh.stdout.read()
		cmdh.stdout.close()
		for line in gpgver.split('\n'):
			if line[0:15] == 'gpg (GnuPG) 2.1':
				return True
	except Exception:
		pass
	return False

# EOF
